-- author: msx80
-- desc:   An arcade racing game!
-- script: lua

-- Thanks to @Fubuki for the awesome
-- car sprite and engine sound!
-- Thanks to the folks on the discord 
-- server for help and suggestions!

-- Want to make a game like this?
-- Read my tutorial on Making a
-- Racing game here:
-- https://github.com/nesbox/TIC-80/wiki/Fun-With-Vectors

-- https://twitter.com/msx80

-- Sitio para jugar el juego en TIC80
-- https://tic80.com/play?cart=573
-- Juegos del mismo programador
-- https://tic80.com/dev?id=1409
local scene4 = {}
require("game_config")
require("helper")
window = require ("zoom-and-move-window")
--require("no_usadas")


t=0
modeStart=0

curLev = nil
-- all entities (cars, tires, etc)
ents={}
-- subset of ents with only the cars
-- this get sorted to calculate cars
-- position and ranking
cars={}

local imageFile
imageFilecar=nil
imagerock=nil
imagetire=nil
imagecarcpu=nil
gameend=false





-- function that returns a "brain"
-- (the object that decide how to move the car)
-- based on the gamepad buttons
function playerBrain(baseButton)
  return function(o)
    --if editIdx==1 then -- if editing, don't move the car
     if love.keyboard.isDown("left") then o.a=angleAdd(o.a,-TURN_RADIUS) end
     if love.keyboard.isDown("right") then o.a=angleAdd(o.a,TURN_RADIUS) end
     if love.keyboard.isDown("z") then
       -- apply an acceleration vector to current velocity
       local x1,y1 = vector(ACCEL_VALUE,o.a)
       o.vx=o.vx+x1
       o.vy=o.vy+y1 
     end
    --end
  end
end




-- function that returns an AI brain,
-- a brain that automatically drives a
-- cpu car
function aiBrain(accelCoeff, offset)
	local nextp=1
	local p = {x=curLev.points[nextp].x,y=curLev.points[nextp].y}
	-- Con esta linea van los CPU al reves
	--if nextp==#curLev.points then nextp = 1 else nextp = nextp + 1 end
	--print("[aiBrain] actual point: " .. nextp)
	return function(o)
	local dst1 = distance(o, p)
		--print("aiBrain FT actual dst1: " .. dst1)
		if dst1 < 20 then
			print("[aiBrain] FT actual dst1: " .. dst1 .. " Level points: " .. #curLev.points)
			--nextp = incw(nextp, #curLev.points)
			if nextp==#curLev.points then nextp = 1 else nextp = nextp + 1 end
			p.x=curLev.points[nextp].x
			p.y=curLev.points[nextp].y
			local offx, offy = vector(offset,curLev.points[nextp].a)
			p.x=p.x+offx
			p.y=p.y+offy
			print("[aiBrain] actual dst1: " .. dst1 .. " point: " .. nextp)
			print("[aiBrain] proximo point: " .. nextp)
		else
		end
		-- get desired angle, toward waypoint
		local desired = angle2(o.x, o.y, p.x, p.y)
		local sgn = angleDir(o.a, desired)
		o.a = angleAdd(o.a,TURN_RADIUS*sgn)
		if true then -- always accelerating!
			-- apply an acceleration vector to current velocity
			local x1,y1 = vector(ACCEL_VALUE*accelCoeff,o.a)
			o.vx=o.vx+x1
			o.vy=o.vy+y1 
		end
	end
end


function adjustColl(a, b)
	if a.fixed and b.fixed then return end
	local dist = distance(a,b)
	local pen = a.r+b.r-dist
	if pen>0 then
		-- collision! find the angle pointing from one center to the other 
		print("[adjustColl] Collision: " )
		local an = angle2(a.x, a.y, b.x, b.y)
		-- first resolve the penetration by pushing apart the entities 
		-- by half the penetration each
		local dx,dy = vector(pen, an)
		-- then we add a small impulse to the velocity in the same direction
		-- for some bounciness
		local impact = vecLen(a.vx-b.vx,a.vy-b.vy) -- difference between velocities  
		local dvx,dvy = vector(impact*0.8, an) --0.8 is empirical, 1 was too bouncy

		if not a.fixed then
			a.x = a.x-dx
			a.y = a.y-dy
			a.vx = a.vx-dvx
			a.vy = a.vy-dvy
		end

		if not b.fixed then
			b.x = b.x+dx
			b.y = b.y+dy
			b.vx = b.vx+dvx
			b.vy = b.vy+dvy
		end
	end
end

function loadEntities(lvl, humans, cpus, laps)
	curLev = lvl
	curLev.laps = laps
	curLev.humans = humans
	ents={}
	cars={}
 
	for x=0,29 do
		for y=0,16 do
			-- Aqui es donde se cargarian los tiles
			--mset(x,y,mget(x+curLev.mx, y+curLev.my))
		end
	end
 
 
	for i=1, humans do
		local c = ins(ents, {
			a=0,
			vx=0,
			vy=0,
			r=10,
			spr=256+ #ents*16,
			name="P"..i,
			human=true,
			lastCheckTime=0,
			brain=playerBrain(0+(i-1)*8),
			directed=true,
			car=true,
			lap=0,
			nextCheck=#(curLev.checks)
		});
		ins(cars, c)
	end
	for i=1, cpus do
		local r= love.math.random(1,#cpuLevels)
		local c = ins(ents, {
			a=0,
			vx=0,
			vy=0,
			r=10,
			spr=256+ #ents*16,
			name="CPU",
			human=false,
			lastCheckTime=0,
			brain=aiBrain(cpuLevels[r].accelCoeff,
			cpuLevels[r].offset),
			directed=true,
			car=true,
			lap=0,
			nextCheck=#(curLev.checks),
		});
		ins(cars, c)
	end
 
	--place cars
	for i=1,#cars do
		cars[i].x = 58+ 20*((i-1) % 2)
		cars[i].y = 230+ 20*math.floor((i-1) / 2)
	end

	 -- give initial sorting
	 sortCars()
	-- Insertar obstaculos, ruedas y piedras
	for _,e in pairs(lvl.ents) do
		local o = {x=e.x, y=e.y, spr=e.spr, a=0, vx=0, vy=0, r=3}
		if e.spr==496 then -- tire
			o.r=3
		elseif e.spr==497 then -- fix
			o.fixed=true
			o.r=4
		end
		ins(ents, o)
	end
end

function lapTime(lapStart)
	print("[lapTime]" )
	return roundDec( (love.timer.getTime( )-lapStart) / 600, 3)
end

function lapCompleted(o)
	o.nextCheck = 1
	o.lap=o.lap+1
	print("[lapCompleted] lap: " .. o.lap)
	if o.lapStart then
		o.prevLap = lapTime(o.lapStart)
		if o.bestLap then
			if o.prevLap < o.bestLap then o.bestLap = o.prevLap end;
		else
			-- no best lap yet, save this
			o.bestLap = o.prevLap
		end
	end;
	o.lapStart=love.timer.getTime( )
	if (o.lap-1)==curLev.laps then 
		-- this car completed the race. Remove pilot
		o.brain=nil
		-- if no human cars are left, end the game (no need to wait for cpu)
		for i=1,#cars do
			if cars[i].human and cars[i].brain then return end;
		end
		print("[lapCompleted] ENDGAME!!! " )
		--switchMode(MODES.ENDGAME, 0)
		--self.setScene("scene5")
		gameend=true
	end
end

function checkLaps(o)
	local check = curLev.checks[o.nextCheck ]
	if pointInRect(o, check) then
		if o.nextCheck == #curLev.checks then
			print("[checkLaps] Lap completed ")
			lapCompleted(o)
		else
			print("[checkLaps] check passed... [" .. o.nextCheck .. "]" )
			o.nextCheck = o.nextCheck +1
		end
		-- if passed any check, resort
		o.lastCheckTime = love.timer.getTime( )
		--sortCars()
	end
end

--mode=nil


function timeString(o, numPlayer, includeBest)
  local displayTime
  if o.lapStart then
   
   if o.prevLap and (love.timer.getTime( )-o.lapStart) < 1000 then
     -- stick on the current lap for some seconds to show it
     displayTime = o.prevLap
   else
     displayTime = rpad(tostring(lapTime(o.lapStart)), 6, '0')
   end
  else
   displayTime = ""
  end
  
  local bestLap = ""
  if o.bestLap and includeBest then
    bestLap = " Best: "..rpad(tostring(o.bestLap), 6, '0')
  end 
  
  local lap = o.lap
  if lap < 1 then lap = 1 elseif lap>curLev.laps then lap=curLev.laps end
  print("P"..numPlayer..": "..lap.."/"..curLev.laps.." #"..o.pos.." "..displayTime..""..bestLap)
  return "P"..numPlayer..": "..lap.."/"..curLev.laps.." #"..o.pos.." "..displayTime..""..bestLap
end


  
function sortCars()
table.sort(cars, function(a,b)
 local aa=a.lap*100+a.nextCheck
 local bb=b.lap*100+b.nextCheck
 if aa==bb then
   -- same loop and check, see who arrived first
   return a.lastCheckTime < b.lastCheckTime
 else
   return aa>bb
 end
end)
for i=1,#cars do
 cars[i].pos=i
end
end



function scene4:load()
	gameend=false
	width, height = love.graphics.getDimensions( )
	window:load()
	-- Imagen de fondo
	--imageFile = love.graphics.newImage("img/montemarco.png")
	imageFile =     love.graphics.newImage("img/carretera2.png")
	imageFilecar =  love.graphics.newImage("img/car.png")
	imagerock =     love.graphics.newImage("img/rock_1.png")
	imagetire =     love.graphics.newImage("img/tire_1.png")
	--imagecarcpu =   love.graphics.newImage("img/car_red.png")
	imagecarcpu =   love.graphics.newImage("img/car_blue.png")

	-- Cargar pista 1 lvls[1]
	-- players 1
	-- cpu 1
	-- vueltas 3
	loadEntities(lvls[1],1,3,2)
end
function scene4:update(dt)
	t=t+1
	--updateGame()
	for i=1,#ents do
		local o = ents[i]
		if o.brain then o:brain() end
		-- decide and apply friction
		--local inDirt = isInDirt(o)
		local inDirt = 0
		--local fr = btn(5) and BRAKE_FRICTION or (inDirt and DIRT_FRICTION or FRICTION)
		local fr = BRAKE_FRICTION or (inDirt and DIRT_FRICTION or FRICTION)

		-- slow current velocity
		o.vx = o.vx * fr
		o.vy = o.vy * fr
		if o.car then
			-- even if we are not accelerating, we must rotate the
			-- velocity vector to slowly catch up with the direction
			-- vector, simulating the wheels regaining traction.
			-- without this, in absence of acceleration the car would 
			-- spin freely without affecting the direction of movement
			local v = angle(o.vx, o.vy)
			local d = angleDir(v, o.a) 
			o.vx, o.vy = rotate(o.vx, o.vy, ALIGNEMENT * d)

		end
		-- apply velocity to position
		o.x=o.x+o.vx
		o.y=o.y+o.vy
	end
	for i=1,#ents do
		for j=i+1,#ents do
			adjustColl(ents[i], ents[j])
		end
	end
	for i=1,#ents do
		local o = ents[i]
		--if o.x<0 then o.x=0 elseif o.x>800 then o.x=800 end
		if o.x<0 then o.x=0 elseif o.x>width then o.x=width end
		--if o.y<0 then o.y=0 elseif o.y>600 then o.y=600 end
		if o.y<0 then o.y=0 elseif o.y>height then o.y=height end
	end
	for i=1,#cars do
		checkLaps(cars[i])
	end
	
	for i=1, curLev.humans do
		-- human players are always the first entities
		pp(timeString(ents[i], i, curLev.humans == 1), 1+(i-1)*120, 129)
		-- only two will ever show but hey
	end
	window:update(dt)
end
function scene4:draw()
	love.graphics.setColor(1, 1, 1)
	window:draw()
	-- Draw fondo
	love.graphics.draw(imageFile,0,0,0,0.20,0.20)
	if gameend then
		love.graphics.print("GAME END\nPress e to another race!", 100, 200)
	end

	for i=1,#ents do
		local o = ents[i]
		if DEBUG then
			--love.graphics.line(o.x,o.y, o.a , 10)
			love.graphics.setColor(1, 1, 0)
			love.graphics.line(o.x,o.y, o.x+o.vx*20,o.y+o.vy*20)
			-- Render de los objetos del juego
			local idx = rou(o.a*16/pi2) % 16 
			spr(o.spr + (idx % 4),flr(o.x)-4, flr(o.y)-4,3, 1,0, math.floor(idx / 4))
			love.graphics.setColor(1, 0, 1)
			love.graphics.circle("fill", o.x,o.y, o.r)
			love.graphics.setColor(0, 1, 1)
			love.graphics.rectangle("fill", flr(o.x), flr(o.y),1,1)
			if o.car then
				-- Render de los objetos del juego
				spr(o.spr + (idx % 4),flr(o.x)-4, flr(o.y)-4,3, 1,0, math.floor(idx / 4))
				love.graphics.print(o.lap.."/"..o.nextCheck, o.x+4, o.y+4)
			end
		end
		if o.car then
			if o.human then
				--spr2(flr(o.x)-4, flr(o.y)-4,o.a)
				love.graphics.draw(imageFilecar,flr(o.x)-4,flr(o.y)-4,o.a,1,1,8,11)
			else
				love.graphics.draw(imagecarcpu,flr(o.x)-4,flr(o.y)-4,o.a,1,1,8,11)
			end
		else
			if o.fixed then
				love.graphics.draw(imagerock,flr(o.x)-4,flr(o.y)-4,o.a,1.2,1.2)
			else
				love.graphics.draw(imagetire,flr(o.x)-4,flr(o.y)-4,o.a,1.2,1.2)
			end
		end
		

	end
	-- Get mpuse position
	local x,y = love.mouse.getPosition()
	love.graphics.print("Mouse x: " .. x .. " y: " .. y,150,150)
	if DEBUG then
		-- Dibujar cosas del circuito
		for _,e in pairs(lvls[1].points) do
			--local o = {x=e.x, y=e.y}
			--love.graphics.rectangle("fill", e.x,e.y, 10,10)
			love.graphics.setColor(1, 0, 0)
			love.graphics.circle("line", e.x,e.y, 12)			
		end
		for _,e in pairs(lvls[1].checks) do
			--local o = {x=e.x, y=e.y}
			love.graphics.setColor(0, 0, 1)
			love.graphics.rectangle("line", e.x,e.y, e.w,e.h)		
		end
		-- Fin dibujitos circuito
	end
end

function scene4:wheelmoved(x, y)
	window:wheelmoved(x, y)
end

function scene4:keypressed(charCode, keyCode)
	if love.keyboard.isDown("d")    then
		if DEBUG then
			DEBUG = false
		else
			DEBUG = true
		end
	end
	if gameend and love.keyboard.isDown("e") then
		gameend=false
		self.setScene("scene5")
	end
	return nil
end

function scene4:mousepressed()

end

function scene4:mousemoved (x, y)
	--love.graphics.print("Mouse x: " .. x .. " y: " .. y,150,150)
end

return scene4
